"""
Generate a synthetic photoelastic response.
"""
import numpy as np

import numba

from pepe.preprocess import circularMask

# fastmath relaxes some of the floating point precision,
# and leads to an accumulated error of around 1e-10
# It doesn't speed up a ton, but this error is so small that
# any amount of speed up is worth it (I think it's about 5ms/evaluation)
@numba.njit(parallel=False, fastmath=True)
def genSyntheticResponse(forceArr, alphaArr, betaArr, fSigma, radius, pxPerMeter=1, brightfield=True, imageSize=None, center=None, mask=None):
    """
    Generate the theoretical photoelastic response given the material
    properties and a set of forces.

    Based on Jonathan Kollmer's PeGS (matlab):
    https://github.com/jekollmer/PEGS

    and Olivier Lantsoght's PeGS_Py:
    https://git.immc.ucl.ac.be/olantsoght/pegs_py

    Parameters
    ----------

    forceArr : np.ndarray[Z]
        Array of force magnitudes acting on the particle (Z is number of contacts
        aka coordination number)

    alphaArr : np.ndarray[Z]
        Array of angles representing force contact angles.

    betaArr : np.ndarray[Z]
        Array of angles representing force contact positions.

    fSigma : float
        Stress optic coefficient, relating to material thickness, wavelength of light
        and other material property (C).

    radius : float
        Radius of the particle that is being simulated in pixels. If pxPerMeter is
        set to None, this value will be assumed to already have been converted to meters.

    pxPerMeter : float
        The number of pixels per meter for the simulated image. If not provided (or set to 1), the radius
        value will be assumed to already have been converted to meters.

    brightfield : bool
        Whether the intensity should be simulated as seen through a brightfield (True)
        polariscope or darkfield (False) polariscope.

    imageSize : tuple(np.int64[2]) or None
        The size of the canvas the stress pattern will be drawn on. Set to None to automatically
        determine this based on the radius.

    center : np.float64[2] or None
        Where the center of the particle should be position in on the image canvas. Useful for
        creating a composite image of many particles, as the final image can be made just by
        adding each individual matrix together. Set to None to center the particle automatically.

    mask : np.uint8[H,W]
        Array of values representing which points, within the circle, to calculate the intensity
        for. Points with non-zero value represented desired points, and locations with value of
        0 will not be calculated.

        Points with value of 1 outside of the circular particle will be ignored.

    """

    # Deal with parameters about the image
    # Note that it is very helpful to have an odd size, such that there is
    # a central pixel
    if imageSize is None:
        # 11 is arbitrary, just a small, odd number
        imageSize = (np.int64(radius*2)+11, np.int64(radius*2)+11)

    if center is None:
        center = np.array([imageSize[0]/2, imageSize[1]/2], dtype=np.float64)

    if mask is None:
        mask = circularMask(imageSize, center, radius)[:,:,0]

    # Create a mask representing the particle, to find the points we need
    # to simulate
    particleMask = circularMask(imageSize, center, radius)[:,:,0]
    # Grab every point inside the circle

    # Normally we could just do this:
    #points = np.transpose(np.where(cMask > 0))
    # but numba doesn't quite like this way, so we have to be 
    # a little more creative
    fullMask = particleMask + mask
    whereIndices = np.where(fullMask > 1)
    points = np.zeros((len(whereIndices[0]), 2), dtype=np.int64)
    # There is a chance that these indices are backwards, but
    # because we have rotational symmetry, it doesn't really matter...
    # BUT if there is ever some weird anisotropy bug or something,
    # try switching these indices
    points[:,0] = whereIndices[0]
    points[:,1] = whereIndices[1]

    intensityArr = np.zeros((imageSize[0], imageSize[1]))

    # As per conventions through the rest of the library, order of points is y,x
    for i in numba.prange(points.shape[0]):
        # While we index for the points on the entire grid,
        # we have to pass the stress evaluation method the position relative
        # to the center of the partice
        intensityArr[points[i,0],points[i,1]] = evaluateStress(points[i,0]-center[0], points[i,1]-center[1], forceArr, alphaArr, betaArr, fSigma, radius, pxPerMeter, brightfield)

    return intensityArr
    

# See genSyntheticResponse note about fastmath
@numba.njit(fastmath=True)
def evaluateStress(yInd, xInd, forceArr, alphaArr, betaArr, fSigma, radius, pxPerMeter, brightfield=True):
    """
    Evaluate the stress (and resulting photoelastic intensity) at a particular point within
    a particle subject to a set of forces. 

    No checks are done to insure that the point provided is within the particle itself -- this
    should be done externally.

    Parameters
    ----------

    yInd : int
        The indexed y of the point we are computing the stress for, relative
        to the center of the particle.

    xInd : int
        The indexed x of the point we are computing the stress for, relative
        to the center of the particle.

    forceArr : np.ndarray[Z]
        Array of force magnitudes acting on the particle (Z is number of contacts,
        aka coordination number).

    alphaArr : np.ndarray[Z]
        Array of angles representing force contact angles.

    betaArr : np.ndarray[Z]
        Array of angles representing force contact positions.

    fSigma : float
        Stress optic coefficient, relating to material thickness, wavelength of light
        and other material property (C).

    radius : float
        Radius of the particle that is being simulated in pixels. If pxPerMeter is
        set to None, this value will be assumed to already have been converted to meters.

    pxPerMeter : float
        The number of pixels per meter for the simulated image. While not recommended,
        set value to 1 if the positions and radius have already been converted to meters.

    brightfield : bool
        Whether the intensity should be simulated as seen through a brightfield (True)
        polariscope or darkfield (False) polariscope.
        
    """
    # We expect that distance/position values are NOT in physical units yet
    # so we have to divide by pxPerMeter to get real values
    xM = xInd/pxPerMeter
    yM = yInd/pxPerMeter
    radiusM = radius/pxPerMeter

    # Coordination number
    z = len(forceArr)

    # Iterate over each force, and calculate the total stress
    # attributed to that force (at the given point)
    sigmaXX, sigmaXY, sigmaYY = 0,0,0
    for i in range(z):
        # Adjust beta for rotation to match the real images we are working with
        b1 = -betaArr[i] + np.pi
        # The middle expression is really just the sign function, but we
        # need sign(0)=1 (sign(0)=0 by default) so we do a little
        # boolean magic
        b2 = b1 - (-1 + 2*int(np.sign(alphaArr[i]) > -1)) * np.pi + 2*alphaArr[i]

        # Calculate the length of the chord
        chordX = radiusM * (np.sin(b2) - np.sin(b1))
        chordY = radiusM * (np.cos(b2) - np.cos(b1))
        chordLength = np.sqrt(chordX**2 + chordY**2)

        # Calculate the vector for the given position to the chord
        vecX = xM - radiusM*np.sin(b1)
        vecY = -yM - radiusM*np.cos(b1) # Not sure why there is negative here
        rVec = np.sqrt(vecX**2 + vecY**2)

        # Error check
        # See if either the vector or the chord length is zero
        if rVec*chordLength == 0.:
            #print("Bad length")
            return 0.

        # Normalize the components of the chord
        chordX /= chordLength
        chordY /= chordLength

        arccosArgument = (vecX*chordX + vecY*chordY) / rVec

        # At the very center, this argument will be almost exactly 1,
        # but we'll get a nan if it is exactly 1, so we need to subtract a little
        # bit slightly above the order of machine error
        # This is most likely the issue that Jonathan commented about in his code
        #arccosArgument -= np.float64(xInd+yInd + xInd*yInd == 0)*1e-10
        arccosArgument -= 1e-10

        th = np.sign(vecY*chordX - vecX*chordY) * np.arccos(arccosArgument)

        s1 = -2/np.pi * forceArr[i]*arccosArgument / rVec
        s2 = -1/(np.pi*radiusM) * forceArr[i] * (-np.sin(alphaArr[i]))
        # Update: This was changed from s1 - s2 to just s1 thanks to
        # Ben McMillan finding the bug. For more information, seeA
        # https://github.com/jekollmer/PEGS/pull/4
        # So s2 is not actually needed anymore, but I'll leave it in just in case...
        sr = s1

        th += betaArr[i] - np.pi/2 - alphaArr[i]

        sigmaXX += sr*np.sin(th)**2
        sigmaYY += sr*np.cos(th)**2
        sigmaXY += 0.5*s1*np.sin(2*th)

    pressureDiff = sigmaXX - sigmaYY
    principleStressDiff = np.sqrt(pressureDiff**2 + 4*sigmaXY**2)

    # Brightfield vs darkfield we use cos vs sin
    if brightfield:
        return np.cos(np.pi/fSigma * principleStressDiff)**2
    else:
        return np.sin(np.pi/fSigma * principleStressDiff)**2



"""
Implementations of network analyses, like contact networks or adjacency matrices.
"""

import numpy as np

from sklearn.neighbors import KDTree 

from pepe.preprocess import circularMask
from pepe.analysis import gSquared

def adjacencyMatrix(centers, radii, contactPadding=5, neighborEvaluations=6):
    """
    Calculate the (unweighted) adjacency matrix, aka neighbor matrix,
    of a set of particles (centers and radii).

    For finding centers and radii, see `pepe.tracking`.

    Optimized for large numbers of particles by using a kd-tree to only
    evaluate potential contacts for nearby particles.

    Parameters
    ----------

    centers : np.ndarray[N,2] 
        A list of N centers of format [y, x].

    radii : np.ndarray[N]
        A list of N radii, corresponding to each particle center

    contactPadding : int
        Maximum difference between distance and sum of radii for which
        two particles will still be considered in contact.

    neightborEvaluations : int
        How many of the closest points to find via the kd tree and test
        as potential contacts. For homogeneous circles, or approximately
        homogenous (< 2:1 size ratio), 6 should be plenty.


    Returns
    -------

    adjMat : np.ndarray[N,N]
        Unweighted adjacency matrix
    """

    adjMatrix = np.zeros([len(centers), len(centers)])
    # Instead of going over every point and every other point, we can just take nearest neighbors, since
    # only neighbor particles can be in contact
    kdTree = KDTree(centers, leaf_size=10)
    
    # In 2D, 8 neighbors should be more than enough
    # +1 is so we can remove the actual point itself
    # Though if we have very few points, we may not even have 9 total
    dist, ind = kdTree.query(centers, k=min(neighborEvaluations+1, len(centers)))
   
    # See if the distance between nearby particles is less than the
    # sum of radii (+ the padding)
    for i in range(len(centers)):
        for j in range(len(ind[i])):
            if radii[i] + radii[ind[i][j]] + contactPadding > dist[i][j]: 
                adjMatrix[i][ind[i][j]] = 1
                adjMatrix[ind[i][j]][i] = 1

    return adjMatrix


def weightedAdjacencyMatrix(photoelasticSingleChannel, centers, radii, contactPadding=5, g2MaskPadding=1, contactThreshold=.1, neighborEvaluations=6):
    """
    Calculate a weighted adjacency matrix of a system of particles
    based on the average G^2 of each particle.

    For finding centers and radii, see `pepe.tracking`.

    For unweighted adjacency matrix, see `pepe.analysis.adjacencyMatrix()`.

    Parameters
    ----------

    photelasticSingleChannel : np.uint8[H,W]
        A single channel image in array form of the photoelastic response of particles.

    centers : np.ndarray[N,2] 
        A list of N centers of format [y, x].

    radii : np.ndarray[N]
        A list of N radii, corresponding to each particle center

    contactPadding : int
        Maximum difference between distance and sum of radii for which
        two particles will still be considered in contact.

    g2MaskPadding : int
        Number of pixels to ignore at the edge of each particle when calculating the average
        G^2.

    contactThreshold : float
        The neighbor weight value under which edges will be removed from the network,
        as they would be considered to weak to represent anything physical. This will
        help remove particle neighbors that are only barely touching, but not transmitting
        any real force.

    neightborEvaluations : int
        How many of the closest points to find via the kd tree and test
        as potential contacts. For homogeneous circles, or approximately
        homogenous (< 2:1 size ratio), 6 should be plenty.


    Returns
    -------

    adjMat : np.ndarray[N,N]
        Weighted adjacency matrix
    """

    # First, we calculate the unweighted network
    unweightedAdjMat = adjacencyMatrix(centers, radii, contactPadding, neighborEvaluations)

    # Calculate g^2 for entire image
    gSqr = gSquared(photoelasticSingleChannel)

    particleGSqrArr = np.zeros(len(centers))
    # Apply a circular mask to each particle, and sum the gradient squared
    # for each one
    for i in range(len(centers)):
        mask = circularMask(photoelasticSingleChannel.shape, centers[i], radii[i] - g2MaskPadding, channels=None)
        # Divide by radius squared, since we could have particles of different sizes
        particleGSqrArr[i] = np.sum(gSqr * mask) / np.sum(mask)

    # Outer product, so average gradients are all multiplied together
    weightedAdjMat = np.multiply.outer(particleGSqrArr, particleGSqrArr) * unweightedAdjMat

    # Clean up the matrix by normalizing the diagonals
    maxValue = np.max(weightedAdjMat)
    for i in range(len(centers)):
        weightedAdjMat[i,i] = maxValue

    weightedAdjMat /= np.max(weightedAdjMat)

    # And remove all edges below the cutoff threshold
    weightedAdjMat[weightedAdjMat < contactThreshold] = 0

    return weightedAdjMat




import numpy as np

def derivative(xArr, dx=1):
    """
    Use five-point central difference stencil to calculate the discrete derivative
    of a set of data.
    """
    dArr = np.zeros(len(xArr))

    # First and last points have to use a directional stencil
    dArr[0] = (xArr[0] - xArr[1]) / dx
    dArr[-1] = (xArr[-2] - xArr[-1]) / dx

    dArr[1] = (xArr[0] - xArr[2]) / (2*dx)
    dArr[-2] = (xArr[-3] - xArr[-1]) / (2*dx)

    # Now everything in between
    dArr[2:-2] = (xArr[:-4] - 8*xArr[1:-3] + 8*xArr[3:-1] - xArr[4:]) / (12*dx)

    return dArr



"""
Defines the Trial object to assist with analyzing the output from `pepe.auto.forceSolve()`.
"""

import numpy as np
import inspect
import pickle
import os
from PIL import Image

from pepe.auto import forceSolveArgDTypes
from pepe.analysis import forceOptimizeArgDTypes
from pepe.tracking import circleTrackArgDTypes
from pepe.utils import parseList
from pepe.preprocess import checkImageType

class Trial():

    def __init__(self, datasetDir=None, fArr=None, aArr=None, bArr=None, cArr=None, rArr=None, thArr=None):
        """
        Helper class for reading in the output of `pepe.auto.forceSolve()`.

        Parameters
        ----------

        datasetDir : str
            Path to the directory containing readme and pickle files.

        fArr : array_like or None
            Force array, if Trial object is to be constructed manually.

        aArr : array_like or None
            Alpha array, if Trial object is to be constructed manually.

        bArr : array_like or None
            Beta array, if Trial object is to be constructed manually.

        cArr : array_like or None
            Center array, if Trial object is to be constructed manually.

        rArr : array_like or None
            Radius array, if Trial object is to be constructed manually.

        thArr : array_like or None
            Angle array, if Trial object is to be constructed manually.
        """

        if datasetDir is not None:
            self.datasetPath = os.path.abspath(datasetDir) + '/'

            if not os.path.exists(self.datasetPath):
                raise Exception(f'Directory {self.datasetPath} does not exist!')

            # Check to make sure all of the pickle files exist
            fileNames = ['forces', 'alphas', 'betas', 'centers', 'radii', 'angles']
            fileExists = {}
            for f in fileNames:
                fileExists[f] = os.path.exists(self.datasetPath + f + '.pickle')

            # Make sure we have at least one of the files
            if np.sum(np.int64(list(fileExists.values()))) == 0:
                raise Exception(f'Directory {self.datasetPath} doesn\'t contain any data... Are you sure this is the correct directory?')

            # If we are only missing one or two, we can just print that
            for f in fileNames:
                if not fileExists[f]:
                    print(f'Warning: dataset {datasetDir} is missing {f}.pickle file!\n {f[0].upper()}{f[1:]} will not be available for this object.')

            # Load in all of the pickle files
            if fileExists['forces']:
                with open(self.datasetPath + 'forces.pickle', 'rb') as f:
                    self.forceArr = pickle.load(f)
            else:
                self.forceArr = None

            if fileExists['betas']:
                with open(self.datasetPath + 'betas.pickle', 'rb') as f:
                    self.betaArr = pickle.load(f)
            else:
                self.betaArr = None

            if fileExists['alphas']:
                with open(self.datasetPath + 'alphas.pickle', 'rb') as f:
                    self.alphaArr = pickle.load(f)
            else:
                self.alphaArr = None

            if fileExists['centers']:
                with open(self.datasetPath + 'centers.pickle', 'rb') as f:
                    self.centerArr = pickle.load(f)
            else:
                self.centerArr = None

            if fileExists['radii']:
                with open(self.datasetPath + 'radii.pickle', 'rb') as f:
                    self.radiusArr = pickle.load(f)
            else:
                self.radiusArr = None

            if fileExists['angles']:
                with open(self.datasetPath + 'angles.pickle', 'rb') as f:
                    self.angleArr = pickle.load(f)
            else:
                self.angleArr = None

            # And try to load in the settings
            settingsFile = self.datasetPath + 'readme.txt'
            self.settings = {}
            settingDTypes = forceSolveArgDTypes.copy()
            settingDTypes.update(circleTrackArgDTypes)
            settingDTypes.update(forceOptimizeArgDTypes)

            if os.path.exists(settingsFile):
                    fileObj = open(settingsFile, 'r')
                    for line in fileObj:
                        # Check each line and see if it looks like a dictionary value
                        split = line.split(':')
                        # Read settings into the master settings file
                        if len(split) == 2 and split[0].strip() in settingDTypes.keys():
                            # Cast to the type of the value already in the dict
                            if split[1].strip() == 'None':
                                self.settings[split[0].strip()] = None
                            else:
                                if '[' in split[1]:
                                    self.settings[split[0].strip()] = parseList(split[1].strip(), dtype=settingDTypes[split[0].strip()])
                                else:
                                    # Bools need a special condition
                                    if settingDTypes[split[0].strip()] is bool:
                                        val = split[1].strip() == 'True'
                                    else:
                                        val = settingDTypes[split[0].strip()](split[1].strip())

                                    self.settings[split[0].strip()] = val


            # Load in the particle identities too, since that is often very useful
            if 'genFitReport' in self.settings.keys() and self.settings["genFitReport"]:
                self.identitiesImage = Image.fromarray(checkImageType(self.datasetPath + 'FitReport_src/particle_identities.png'))
            else:
                self.identitiesImage = None
        else:
            # If we are explicitly passed arrays, we just assign those
            self.forceArr = fArr
            self.alphaArr = aArr
            self.betaArr = bArr
            self.centerArr = cArr
            self.radiusArr = rArr
            self.angleArr = thArr

            self.identitiesImage = None
            self.datasetPath = None

        # Assign some constants
        if self.forceArr is not None:
            self.numParticles = len(self.forceArr)
            self.numForces = [len(f) for f in self.forceArr]
            self.numTimesteps = [f[0].shape[-1] for f in self.forceArr if len(f) > 0][0]

    def fixAllNan(self, interpolateForce=True):
        """
        Replaces all nan values in all data arrays with physically inspired
        values.

        Enacts the following behavior on each data type:

            forceArr: replaces with 0 or interpolates
            angleArr: interpolates
            alphaArr: interpolates
            betaArr: interpolates
            radiusArr: interpolates
            centerArr: interpolates
        """
        for i in range(self.numParticles):

            if not interpolateForce:
                # Forces
                # Just replace all nan values with 0
                self.forceArr[i] = np.where(np.isnan(self.forceArr[i]), 0, self.forceArr[i])

            # Interpolate for the rest
            # In theory, when one is nan, the others should be as well, but
            # this isn't guaranteed, so it is safest just to go through each
            # separately

            # Angles
            angleNanIndices = np.where(np.isnan(self.angleArr[i]))[0]

            for j in range(len(angleNanIndices)):
                previousGoodPoint = [ind for ind in np.arange(angleNanIndices[j]) if not ind in angleNanIndices]
                previousGoodPoint = previousGoodPoint[-1] if len(previousGoodPoint) > 0 else -1
                nextGoodPoint = [ind for ind in np.arange(angleNanIndices[j], self.numTimesteps) if not ind in angleNanIndices]
                nextGoodPoint = nextGoodPoint[0] if len(nextGoodPoint) > 0 else -1
               
                # If one of the points before or after doesn't exist, just copy over the
                # one that does. We should never have a case in which both points don't
                # exist, because that would mean we have an entire array of nan values
                if previousGoodPoint < 0:
                    self.angleArr[i][angleNanIndices[j]] = self.angleArr[i][nextGoodPoint]
                elif nextGoodPoint < 0:
                    self.angleArr[i][angleNanIndices[j]] = self.angleArr[i][previousGoodPoint]
                else:
                    self.angleArr[i][angleNanIndices[j]] = (self.angleArr[i][previousGoodPoint] * (angleNanIndices[j] - previousGoodPoint) + self.angleArr[i][nextGoodPoint] * (nextGoodPoint - angleNanIndices[j])) / (nextGoodPoint - previousGoodPoint)


            # Centers
            centerNanIndices = np.where(np.isnan(self.centerArr[i][:,0]))[0]

            for j in range(len(centerNanIndices)):
                previousGoodPoint = [ind for ind in np.arange(centerNanIndices[j]) if not ind in centerNanIndices]
                previousGoodPoint = previousGoodPoint[-1] if len(previousGoodPoint) > 0 else -1
                nextGoodPoint = [ind for ind in np.arange(centerNanIndices[j], self.numTimesteps) if not ind in centerNanIndices]
                nextGoodPoint = nextGoodPoint[0] if len(nextGoodPoint) > 0 else -1
              
                # If one of the points before or after doesn't exist, just copy over the
                # one that does. We should never have a case in which both points don't
                # exist, because that would mean we have an entire array of nan values
                if previousGoodPoint < 0:
                    self.centerArr[i][centerNanIndices[j]] = self.centerArr[i][nextGoodPoint]
                elif nextGoodPoint < 0:
                    self.centerArr[i][centerNanIndices[j]] = self.centerArr[i][previousGoodPoint]
                else:
                    # Loop over y and x
                    for k in range(2):
                        self.centerArr[i][angleNanIndices[j]][k] = (self.centerArr[i][previousGoodPoint][k] * (centerNanIndices[j] - previousGoodPoint) + self.centerArr[i][nextGoodPoint][k] * (nextGoodPoint - centerNanIndices[j])) / (nextGoodPoint - previousGoodPoint)

            # Radii
            radiusNanIndices = np.where(np.isnan(self.radiusArr[i]))[0]

            for j in range(len(radiusNanIndices)):
                previousGoodPoint = [ind for ind in np.arange(radiusNanIndices[j]) if not ind in radiusNanIndices]
                previousGoodPoint = previousGoodPoint[-1] if len(previousGoodPoint) > 0 else -1
                nextGoodPoint = [ind for ind in np.arange(radiusNanIndices[j], self.numTimesteps) if not ind in radiusNanIndices]
                nextGoodPoint = nextGoodPoint[0] if len(nextGoodPoint) > 0 else -1

                # If one of the points before or after doesn't exist, just copy over the
                # one that does. We should never have a case in which both points don't
                # exist, because that would mean we have an entire array of nan values
                if previousGoodPoint < 0:
                    self.radiusArr[i][radiusNanIndices[j]] = self.radiusArr[i][nextGoodPoint]
                elif nextGoodPoint < 0:
                    self.radiusArr[i][radiusNanIndices[j]] = self.radiusArr[i][previousGoodPoint]
                else:  
                    self.radiusArr[i][angleNanIndices[j]] = (self.radiusArr[i][previousGoodPoint] * (radiusNanIndices[j] - previousGoodPoint) + self.radiusArr[i][nextGoodPoint] * (nextGoodPoint - radiusNanIndices[j])) / (nextGoodPoint - previousGoodPoint)

            # For the last two, we have to iterate over each force
            for j in range(self.numForces[i]):

                # Alphas
                alphaNanIndices = np.where(np.isnan(self.alphaArr[i][j]))[0]

                for k in range(len(alphaNanIndices)):
                    previousGoodPoint = [ind for ind in np.arange(alphaNanIndices[k]) if not ind in alphaNanIndices]
                    previousGoodPoint = previousGoodPoint[-1] if len(previousGoodPoint) > 0 else -1
                    nextGoodPoint = [ind for ind in np.arange(alphaNanIndices[k], self.numTimesteps) if not ind in alphaNanIndices]
                    nextGoodPoint = nextGoodPoint[0] if len(nextGoodPoint) > 0 else -1

                    # If one of the points before or after doesn't exist, just copy over the
                    # one that does. We should never have a case in which both points don't
                    # exist, because that would mean we have an entire array of nan values
                    if previousGoodPoint < 0:
                        self.alphaArr[i][j][alphaNanIndices[k]] = self.alphaArr[i][j][nextGoodPoint]
                    elif nextGoodPoint < 0:
                        self.alphaArr[i][j][alphaNanIndices[k]] = self.alphaArr[i][j][previousGoodPoint]
                    else:   
                        self.alphaArr[i][j][alphaNanIndices[k]] = (self.alphaArr[i][j][previousGoodPoint] * (alphaNanIndices[k] - previousGoodPoint) + self.alphaArr[i][j][nextGoodPoint] * (nextGoodPoint - alphaNanIndices[k])) / (nextGoodPoint - previousGoodPoint)

                # Betas
                betaNanIndices = np.where(np.isnan(self.betaArr[i][j]))[0]

                for k in range(len(betaNanIndices)):
                    previousGoodPoint = [ind for ind in np.arange(betaNanIndices[k]) if not ind in betaNanIndices]
                    previousGoodPoint = previousGoodPoint[-1] if len(previousGoodPoint) > 0 else -1
                    nextGoodPoint = [ind for ind in np.arange(betaNanIndices[k], self.numTimesteps) if not ind in betaNanIndices]
                    nextGoodPoint = nextGoodPoint[0] if len(nextGoodPoint) > 0 else -1
                    
                    # If one of the points before or after doesn't exist, just copy over the
                    # one that does. We should never have a case in which both points don't
                    # exist, because that would mean we have an entire array of nan values
                    if previousGoodPoint < 0:
                        self.betaArr[i][j][betaNanIndices[k]] = self.betaArr[i][j][nextGoodPoint]
                    elif nextGoodPoint < 0:
                        self.betaArr[i][j][betaNanIndices[k]] = self.betaArr[i][j][previousGoodPoint]
                    else:   
                        self.betaArr[i][j][betaNanIndices[k]] = (self.betaArr[i][j][previousGoodPoint] * (alphaNanIndices[k] - previousGoodPoint) + self.alphaArr[i][j][nextGoodPoint] * (nextGoodPoint - alphaNanIndices[k])) / (nextGoodPoint - previousGoodPoint)

                # Forces
                if interpolateForce:
                    forceNanIndices = np.where(np.isnan(self.forceArr[i][j]))[0]

                    for k in range(len(forceNanIndices)):
                        previousGoodPoint = [ind for ind in np.arange(forceNanIndices[k]) if not ind in forceNanIndices]
                        previousGoodPoint = previousGoodPoint[-1] if len(previousGoodPoint) > 0 else -1
                        nextGoodPoint = [ind for ind in np.arange(forceNanIndices[k], self.numTimesteps) if not ind in forceNanIndices]
                        nextGoodPoint = nextGoodPoint[0] if len(nextGoodPoint) > 0 else -1
                        
                        # If one of the points before or after doesn't exist, just copy over the
                        # one that does. We should never have a case in which both points don't
                        # exist, because that would mean we have an entire array of nan values
                        if previousGoodPoint < 0 or nextGoodPoint < 0:
                            self.forceArr[i][j][forceNanIndices[k]] = 0
                        else:   
                            self.forceArr[i][j][forceNanIndices[k]] = (self.forceArr[i][j][previousGoodPoint] * (forceNanIndices[k] - previousGoodPoint) + self.forceArr[i][j][nextGoodPoint] * (nextGoodPoint - forceNanIndices[k])) / (nextGoodPoint - previousGoodPoint)



    def particleNear(self, point, timestep=0):
        """
        Identifies the index of the particle closest to the provided point
        at the provided timestep.
        """

        distances = np.sum((self.centerArr[:,timestep,:] - np.array(point))**2, axis=-1)
        return np.argmin([d for d in distances if not np.isnan(d)])


    def averageForcePositions(self, particleIndex):
        """
        Calculates the time-averaged beta values of forces for the provided
        particle.
        """
        #averageBetaArr = np.array([np.nanmean(self.betaArr[particleIndex][i,:]) for i in range(self.numForces[particleIndex])])
        averageBetaArr = np.array([np.nanmedian(self.betaArr[particleIndex][i,:]) for i in range(self.numForces[particleIndex])])
        # Now correct for periodicity
        #return np.array([ab if ab <= np.pi else ab - 2*np.pi for ab in averageBetaArr])
        return averageBetaArr
                                    

    def __str__(self):
        return self.datasetPath.split('/')[-2] if self.datasetPath is not None else super.__str__()



"""
Filter application and edge detection.
"""
import numpy as np

from scipy.ndimage import filters

import numba
import cv2

def applyFilter(singleChannelFrame, filter):
    """
    Apply an arbitrary filter (in the form of a matrix) to an image.

    Pretty much just a wrapper for `scipy.ndimage.filter.convolve()`.

    Parameters
    ----------

    singleChannelFrame : np.uint8[H,W]
        The frame to perform circle detection on. Should not include multiple
        color channels; if it does contain multiple channels, the first one will
        be used.

    filter : np.ndarray[N,N]
        Square matrix with an odd kernel size (N is odd) to be applied across the image.

    Returns
    -------

    filtedImage : np.uint8[H,W]
        Filtered image.
    """
    if singleChannelFrame.ndim > 2:
        image = singleChannelFrame.astype(np.float64)[:,:,0]
    else:
        image = singleChannelFrame.astype(np.float64)

    # Make sure the filter looks good
    assert filter.shape[0] == filter.shape[1], 'Invalid filter provided; must be square matrix'
    assert filter.shape[0] % 2 == 1, 'Invalid filter provided; must have odd kernel size'

    # So I wrote this and then found out that scipy has what is most likely a 
    # way better implementation, so we'll just use theirs
    
    return filters.convolve(image, filter)

#    kernelSize = filter.shape[0]
#    # Because we always have an odd kernel size
#    halfKernelSize = int((kernelSize - 1) / 2)
#    # Add some padding to our image so we can apply
#    # the filter to the entire image
#    paddedImage = np.zeros((image.shape[0] + 2*halfKernelSize, image.shape[1] + 2*halfKernelSize))
#    paddedImage[halfKernelSize:-halfKernelSize,halfKernelSize:-halfKernelSize] = image
#    
#    return _applyFilterOpt(paddedImage.astype(np.float64), filter, kernelSize).astype(np.uint8)
#    
#@numba.njit()
#def _applyFilterOpt(paddedImage, filter, kernelSize):
#    """
#    numba optimized method that is wrapped by applyFilter().
#    """
#    filteredImage = np.zeros((paddedImage.shape[0] - kernelSize + 1, paddedImage.shape[1] - kernelSize + 1))
#    for i in range(filteredImage.shape[0]):
#        for j in range(filteredImage.shape[1]):
#            filteredImage[i,j] = np.sum(paddedImage[i:i+kernelSize,j:j+kernelSize] * filter)
#
#    return filteredImage


def sobelEdgeDetection(singleChannelFrame, threshold=.005):
    """
    Perform Sobel edge detection on an image, returning a binarized
    version that should help in tracking circles.

    Parameters
    ----------

    singleChannelFrame : np.uint8[H, W]
        The frame to perform circle detection on. Should not include multiple
        color channels; if it does contain multiple channels, the first one will
        be used.

    threshold : float
        Value between 0 and 1 representing the percentage of the maximum image
        intensity under which the binarized image will have a value of 0.

    Returns
    -------

    binarizedImage : np.uint8[H,W]
        Binarized image (containing only values of 0 and 1).
    """

    if singleChannelFrame.ndim > 2:
        image = singleChannelFrame.astype(np.float64)[:,:,0]
    else:
        image = singleChannelFrame.astype(np.float64)

    # sobel filter in x and y direction
    image = filters.sobel(image, 0)**2 + filters.sobel(image, 1)**2
    image -= np.min(image)

    # binarize image
    image = np.uint8(image > np.max(image)*threshold)

    return image


def laplacianEdgeDetection(singleChannelFrame, threshold=.005):
    """
    Perform edge detection on an image using a laplacian kernel,
    returning a version that should help in tracking circles.

    Note that this is actually what we refer to as the 'gradient
    squared' elsewhere in the code, so this is mostly just a
    wrapper for that method.

    Parameters
    ----------

    singleChannelFrame : np.uint8[H, W]
        The frame to perform circle detection on. Should not include multiple
        color channels; if it does contain multiple channels, the first one will
        be used.

    threshold : float
        Value between 0 and 1 representing the percentage of the maximum image
        intensity under which the binarized image will have a value of 0.

    Returns
    -------

    binarizedImage : np.uint8[H,W]
        Binarized image (containing only values of 0 and 1).
    """
    # This has to go here otherwise we get a circular import warning
    # Working on thinking of a more permanent solution rn.
    from pepe.analysis import gSquared

    if singleChannelFrame.ndim > 2:
        image = singleChannelFrame.astype(np.float64)[:,:,0]
    else:
        image = singleChannelFrame.astype(np.float64)

    # Laplacian filter
    image = gSquared(image)
    image -= np.min(image)

    # binarize image
    image = np.uint8(image > np.max(image)*threshold)

    return image


def cannyEdgeDetection(singleChannelFrame, threshold=.005):
    return cv2.Canny(singleChannelFrame, 2*threshold, threshold)




import numpy as np

import pathlib

from pepe.preprocess import checkImageType
from pepe.tracking import convCircle, houghCircle
from pepe.utils import preserveOrderArgsort

DATA_DIR = str(pathlib.Path(__file__).parent.resolve()) + '/test_data'

def test_convCircle_track():
    """
    Test convolutional tracking on generated
    test images.
    """
    for i in range(3):
        image = checkImageType(f'{DATA_DIR}/test_circles_{i+1}.bmp').astype(np.uint8)
        radius = 50
        tolerance = .10 # 10% tolerance of the radius

        # Parameters aren't too important
        particleCenters, particleRadii = convCircle(image, radius, negativeHalo=True,
                                                    intensitySoftmax=1.8, intensitySoftmin=.6, peakDownsample=1,
                                                    offscreenParticles=True, radiusTolerance=2, minPeakPrevalence=.4)

        with open(f'{DATA_DIR}/test_circles_{i+1}_centers.npy', 'rb') as f:
            expectedCenters = np.load(f)

        assert len(expectedCenters) == len(particleCenters), f"Incorrect number of particles detected in test image {i+1}: {len(particleCenters)} vs {len(expectedCenters)}."

        sortOrder = preserveOrderArgsort(particleCenters, expectedCenters)

        difference = np.sqrt(np.sum((particleCenters - expectedCenters[sortOrder])**2, axis=-1))

        percentOfRadius = difference / radius
        assert not True in [p > tolerance for p in percentOfRadius], f"Particle centers not within {tolerance*100}% of expected value in test image {i+1}."



"""
Proximity sort, in which a list of items is sorted to most closely match another list.
"""
import numpy as np
from sklearn.neighbors import KDTree


def preserveOrderSort(oldValues, newValues, padMissingValues=False, maxDistance=None, periodic=False, fillNanSpots=True):
    """
    Order a set of values as similarly to a previous set of values as possible.
    That is, arange `newValues` such that the first element is the one that is
    closest to the first element of `oldValues`, and so on.

    If `padMissingValues=True`, then when `newValues` contains fewer items than
    `oldValues`, elements of `None` will be inserted to preserve the order.

    If `newValues` contains more items than `oldValues`, the novel elements will
    be appended at the end of the array in an arbitrary order.

    If the values are coordinates of the form `newValues=[[x1, y1, z1, ...], [x2, y2, z2, ...], ...]`
    then the proximity will be calculated as simple euclidiean distance.

    Parameters
    ----------

    oldValues : iterable
        A set of values whose order is to be preserved.

    newValues : iterable
        The new values that will be ordered based on proximity to the elements
        of `oldValues`.

    padMissingValues : bool
        Whether or not to leave placeholder values (`None`) in the final result
        for values that do not correspond to any of the original points (True)
        or to only return values that have pairs (False).

    maxDistance : float or None
        The maximum distance a point can be from its old counterpart and still be
        considered the same point. If set to None, every point in the smaller of the
        two arrays (new or old) will always have a partner; if a max distance is set,
        this is no longer guaranteed.

        Not recommended to be used if `padMissingValues=False`, since it will result
        in an array that has an arbitrary sorting order.
        
        eg. `[1, 2, 3]` sorted according to `[1, 3, 5]` would become `[1, None, 3, 2]` but
        then the `None` will be removed, resulting in `[1, 3, 2]` which likely isn't
        useful.

    periodic : bool
        Whether the values to be sorted are periodic such that pi and -pi are actually
        the same point (and should have a separation of 0, not 2pi.

    fillNanSpots : bool
        Whether to fill Nan spots in the old list with unmatched new values (True), or
        to preserve the position of the nan value and append unmatched new values to
        the end (False).

    Returns
    -------

    orderedValues : iterable
        The elements of `newValues` ordered according to proximity to the elements
        of `oldValues`.
    """

    # We already have preserveOrderArgsort written, so we can just wrap that
    order = preserveOrderArgsort(oldValues, newValues, padMissingValues, maxDistance, periodic, fillNanSpots)

    dim = len(newValues[0]) if hasattr(newValues[0], '__iter__') else 1

    # Slightly different behavior for d=1 vs. d>1, since we don't want
    # an extra dimension nested in there if our data is 1d
    if dim == 1:
        return np.array([newValues[i] if i is not None else np.nan for i in order])
    else:
        return np.array([newValues[i] if i is not None else np.repeat(np.nan, dim) for i in order])


def preserveOrderArgsort(oldValues, newValues, padMissingValues=False, maxDistance=None, periodic=False, fillNanSpots=True):
    """
    Return a set of indices that order a set of values as similarly to a previous
    set of values as possible. That is, a set of indices that arranges `newValues`
    such that the first element is the one that is closest to the first element of
    `oldValues`, and so on.

    If `padMissingValues=True`, then when `newValues` contains fewer items than
    `oldValues`, elements of `None` will be inserted to preserve the order.

    If `newValues` contains more items than `oldValues`, the novel elements will
    be appended at the end of the array in an arbitrary order.

    If the values are coordinates of the form `newValues=[[x1, y1, z1, ...], [x2, y2, z2, ...], ...]`
    then the proximity will be calculated as simple euclidiean distance.

    Parameters
    ----------

    oldValues : iterable
        A set of values whose order is to be preserved.

    newValues : iterable
        The new values that will be ordered based on proximity to the elements
        of `oldValues`.

    padMissingValues : bool
        Whether or not to leave placeholder values (`None`) in the final result
        for values that do not correspond to any of the original points (True)
        or to only return values that have pairs (False).

    maxDistance : float or None
        The maximum distance a point can be from its old counterpart and still be
        considered the same point. If set to None, every point in the smaller of the
        two arrays (new or old) will always have a partner; if a max distance is set,
        this is no longer guaranteed.

        Not recommended to be used if `padMissingValues=False`, since it will result
        in an array that has an arbitrary sorting order.
        
        eg. `[1, 2, 3]` sorted according to `[1, 3, 5]` would become `[1, None, 3, 2]` but
        then the `None` will be removed, resulting in `[1, 3, 2]` which likely isn't
        useful.

    periodic : bool
        Whether the values to be sorted are periodic such that pi and -pi are actually
        the same point (and should have a separation of 0, not 2pi.

    fillNanSpots : bool
        Whether to fill Nan spots in the old list with unmatched new values (True), or
        to preserve the position of the nan value and append unmatched new values to
        the end (False).

    Returns
    -------

    orderedIndices : iterable
        Indices corresponding to elements of `newValues` according to proximity
        to the elements of `oldValues`.
    """

    if maxDistance is None:
        maxDistance = np.inf

    # Make sure all of the values are np arrays
    npOldValues = np.array(oldValues)
    npNewValues = np.array(newValues)

    # This is required for the kdtree to work, since it is
    # generalized for higher dimensional spaces
    scalarValues = npOldValues.ndim == 1
    if scalarValues:
        npNewValues = npNewValues[:,None]
        npOldValues = npOldValues[:,None]

    # Convert all potential None values to np.nan (since they play nicer)
    # Numpy can only hold values of None in an array if the dtype is 'object',
    # which is inconvient for working with the kd tree
    # A note, since I have now spent a *very* long time figuring out this bug:
    # np.where(array) will give the indices of all values that are not None/np.nan
    # AND all values that are not 0. so this line used to be
    # ...np.where(npNewValues, npNewValues, np.nan)...
    # but this will convert any values that are exactly 0 to np.nan as wel, which is very bad.
    npNewValues = np.array(np.where(npNewValues != None, npNewValues, np.nan), dtype=np.float64)
    npOldValues = np.array(np.where(npOldValues != None, npOldValues, np.nan), dtype=np.float64)

    # We want to have the longer list of items be the source points
    # for the kdtree, and the shorter be the query points. If they are
    # the same length, it doesn't matter.

    # Detect which values are valid aka not np.nan (see above about nan/None)
    oldValidIndices = np.unique(np.where(np.isnan(npOldValues) == False)[0])
    newValidIndices = np.unique(np.where(np.isnan(npNewValues) == False)[0])

    # If we have an empty old list, we just the original indexing
    if len(oldValidIndices) == 0:
        # If there just aren't any values in the old list
        # then we just return indexing for the new values
        if len(npOldValues) == 0:
            return [i for i in range(len(npNewValues))]
        else:
            # If there are some old values, but they are just all
            # None/nan, then we have to decide what to do based on
            # whether we want to fill in the nan spots or not
            if fillNanSpots:
                # Filling nan spots, we just return the same indexing for the new
                # list, potentially with some None values at the end
                return [i for i in range(len(npNewValues))] + [None]*(len(npOldValues)-len(npNewValues))
            else:
                # Otherwise, we return a similar thing, but with the None values at the
                # beginning, in the same spot as they were for the old list
                return [None]*len(npOldValues) + [i for i in range(len(npNewValues))]

    # If there are no new values, we return empty list (the new values)
    # or a list of None values
    if len(newValidIndices) == 0:
        if padMissingValues:
            return [None]*len(npOldValues)
        else:
            return [None]*len(npNewValues)

    # The list that we will be building
    # We'll append new entries on at the end
    addedIndices = [None for i in range(len(npOldValues))]

    kdTree = KDTree(npOldValues[oldValidIndices], leaf_size=5)
    # This gives us the closest k points to each of the old values
    dist, ind = kdTree.query(npNewValues[newValidIndices], k=len(oldValidIndices))

    # Now convert the indices return from the kd tree (which excluded np.nan values)
    # back to the original order, which includes np.nan values
    ind = [[oldValidIndices[ind[i][j]] for j in range(len(ind[i]))] for i in range(len(ind))]

    if periodic:
        # Points can't be more than pi away from each other on a circle (ignoring circulation
        # direction)
        dist = np.where(dist < np.pi, dist, 2*np.pi - dist)
        # TODO: resort ind since the distances may have changed

    # Now for each new point, grab the closest old point that hasn't already
    # been added. The order we do this in is quite arbitrary (it follows the order
    # the user has passed in the values) which means that if particles move quite
    # large distances each frame, the algorithm may be incorrect. It would be more
    # apt to instead perform a minimization on the sum of the change in particle positions
    # for any possible pair configuration. TODO
    
    # When two new points have the same closest old point, one of them will have to
    # use their second closest old point instead (the farther one). This list records
    # the index their closest neighbor that they are able to use. eg. a value of 0
    # here means that that particular point used it's closest old point, a value of 1
    # means it had to use it's second closest point (because the first closest was
    # already taken).
    indexChoicePositions = [None]*len(newValidIndices)

    # This has to be a while loop with the list here because we may need
    # to recheck certain points (eg. if we first assign a point, and then realize
    # that there is actually a better match later on).
    newIndicesToAssign = list(np.arange(len(newValidIndices)))
    listIndex = 0
    while listIndex < len(newIndicesToAssign):
        # i is the index of the current new point
        i = newIndicesToAssign[listIndex]
        # The indices in possible points index the old array, so when we eventually
        # assign a pair, this index will be the **index** of the actual value. The value
        # will just be i, since that is what is iterating over the new list.

        # Checking to see that the addedIndices array is None is making sure that this new
        # point hasn't already been assigned
        # Check to make sure all of the possible points are within the max distance
        possiblePoints = [ind[i][j] for j in range(len(ind[i])) if dist[i][j] < maxDistance]
        # We don't need to explicitly check if len(possiblePoints) = 0, since the behavior
        # in that case should do nothing, as such points are dealt with at the end

        # We'll likely break out of this loop on the first iteration most of the time,
        # but we need to be able to handle if the first/second/etc. best options are already
        # taken
        for j in range(len(possiblePoints)):
            # It is possible that the closest old point to this new one has already
            # been assigned to another new point.
            if addedIndices[possiblePoints[j]] is None:
                # If it hasn't, of course that's no problem, and we can assign it
                addedIndices[possiblePoints[j]] = i
                indexChoicePositions[i] = j
                break
            else:
                # Otherwise, we have to check to see which is actually closer
                compareNewIndex = addedIndices[possiblePoints[j]]
                if dist[i][j] < dist[compareNewIndex][indexChoicePositions[compareNewIndex]]:
                    # If the new point is closer, we reassign the entry in addedIndices
                    # and change the previous best point to the next best (available) one
                    addedIndices[possiblePoints[j]] = i
                    indexChoicePositions[i] = j
                    
                    # And add the replaced index back into the list so
                    # we can find it a new match
                    indexChoicePositions[compareNewIndex] = None
                    newIndicesToAssign.append(compareNewIndex)
                    break
                else:
                    # If the new point is further, we just continue the j loop and try the
                    # next closest old point
                    continue

        listIndex += 1

    # Check which new indices haven't been assigned yet
    # Again, the values of addedIndices are the new indices, and the indices of
    # addedIndices are the old indices (yikes, am I right?)
    unaddedIndices = [i for i in range(len(newValidIndices)) if not i in addedIndices]

    if fillNanSpots:
        oldNanIndices = np.unique(np.where(np.isnan(npOldValues) == True)[0])
        for i in range(min(len(oldNanIndices), len(unaddedIndices))):
            addedIndices[oldNanIndices[i]] = unaddedIndices[i]

        unaddedIndices = unaddedIndices[len(oldNanIndices):] if len(oldNanIndices) < len(unaddedIndices) else []

    addedIndices += unaddedIndices
    
    if not padMissingValues:
        addedIndices = [i for i in addedIndices if i is not None]

    return addedIndices



"""
Methods to upsample or downsample an image via DNN super-resolution or pyramid scheme, respectively.

Note that the super-resolution models need to be downloaded separately, as they are not
my own original work. Four methods are compatible with this package:

- [EDSR](https://arxiv.org/pdf/1707.02921.pdf) ([implementation](https://github.com/Saafke/EDSR_Tensorflow))

- [ESPCN](https://arxiv.org/abs/1609.05158) ([implementation](https://github.com/fannymonori/TF-ESPCN))

- [FSRCNN](http://mmlab.ie.cuhk.edu.hk/projects/FSRCNN.html) ([implementation](https://github.com/Saafke/FSRCNN_Tensorflow))

- [LapSRN](https://arxiv.org/abs/1710.01992) ([implementation](https://github.com/fannymonori/TF-LapSRN))

The above implementations are simply the ones that I have used in the past; there are certainly others available, which should give comparable results. I am not the creator of any of the above listed implementations either, nor am I associated in any way (other than as a user) with them.

To install any of the above models, you should place the `.pb` weight files into an appropriately named directory within the `preprocess/models/` folder. e.g. directory structure:

```
.../preprocess/
    |-- Rescale.py
    |-- ...
    `-- models/
        |-- FSRCNN/
        |   |-- FSRCNN_x2.pb
        |   |-- FSRCNN_x3.pb
        |   `-- FSRCNN_x4.pb
        `-- LapSRN/
            |-- LapSRN_x2.pb
            |-- LapSRN_x4.pb
            `-- LapSRN_x8.pb
```

Make sure that the naming conventions are consistent: the name of the directory should be the same as the first part of the weight file name, and the second part (separated by an underscore `_`) should be `x` plus the upsample factor. These conventions are important for the package to automatically detect which models are available.

"""

import numpy as np
import os

import cv2
from cv2 import dnn_superres
import pathlib
from skimage.measure import block_reduce

# The models dir should be in a directory called 'models' in the same parent directory
# as this file (likely `pepe/preprocess/models/`)
MODELS_DIR = str(pathlib.Path(__file__).parent.resolve()) + '/models/'

# For the upscale() method, we presumably are only doing a single operation,
# but it is still good practice to hold onto the model we generate. So this
# will be assigned to a proper model after the first call to upscale()
defaultModelInstance = None
defaultModelType = 'FSRCNN'
defaultModelFactor = 2

def modelInstallHelp():
    helpString = f"""
    Note that the super-resolution models need to be downloaded separately, as they are not
    my own original work. Four methods are compatible with this package:

    - [EDSR](https://arxiv.org/pdf/1707.02921.pdf) ([implementation](https://github.com/Saafke/EDSR_Tensorflow))

    - [ESPCN](https://arxiv.org/abs/1609.05158) ([implementation](https://github.com/fannymonori/TF-ESPCN))

    - [FSRCNN](http://mmlab.ie.cuhk.edu.hk/projects/FSRCNN.html) ([implementation](https://github.com/Saafke/FSRCNN_Tensorflow))

    - [LapSRN](https://arxiv.org/abs/1710.01992) ([implementation](https://github.com/fannymonori/TF-LapSRN))

    The above implementations are simply the ones that I have used in the past; there are certainly others available, which should give comparable results. I am not the creator of any of the above listed implementations either, nor am I associated in any way (other than as a user) with them.

    To install any of the above models, you should place the `.pb` weight files into an appropriately named directory within the following folder:
    {MODELS_DIR}.

    e.g. directory structure:

    ```
    .../preprocess/
        |-- Rescale.py
        |-- ...
        `-- models/
            |-- FSRCNN/
            |   |-- FSRCNN_x2.pb
            |   |-- FSRCNN_x3.pb
            |   `-- FSRCNN_x4.pb
            `-- LapSRN/
                |-- LapSRN_x2.pb
                |-- LapSRN_x4.pb
                `-- LapSRN_x8.pb
    ```
    Make sure that the naming conventions are consistent: the name of the directory should be the same as the first part of the weight file name, and the second part (separated by an underscore `_`) should be `x` plus the upsample factor. These conventions are important for the package to automatically detect which models are available.
    """

    print(helpString)

def listSuperResModels(debug=False):

    if not os.path.exists(MODELS_DIR):
        print(f'Warning: models folder ({MODELS_DIR}) does not exist!\nSee pepe.preprocess.modelInstallHelp() for more information about installing models.')
        return {}

    # Determine what models are available
    availableModels = [d for d in os.listdir(MODELS_DIR) if os.path.isdir(MODELS_DIR + d)]
    availableModels = np.sort(availableModels)

    modelFactors = {}

    # Now see which upscale factors are available for each model
    # Each .pb file should be of the form <model_name>_x<upscale_factor>.pb
    for d in availableModels:
        weightFiles = [f for f in os.listdir(MODELS_DIR + d) if len(f) > 2 and f[-2:] == 'pb']

        if len(weightFiles) == 0:
            if debug:
                print(f'Warning: model type \'{d}\' has no valid implementations!')
            continue
        
        availableScalingFactors = []
        # Now extract the scaling factor
        for wF in weightFiles:
            try:
                # Cut off the extension, take the portion after the 'x', and cast to int
                upscaleFactor = int(wF[:-3].split('x')[-1])
                availableScalingFactors.append(upscaleFactor)
            except:
                if debug:
                    print(f'Warning: file \'{d}/{wF}\' does not conform to naming conventions.\nAll files should be named in the format <model_type>_x<upscale_factor>.pb!\nSee pepe.preprocess.modelInstallHelp() for more information about installing models.')

        modelFactors[d] = list(np.sort(availableScalingFactors))

    return modelFactors


def prepareSuperResModel(modelName, upscaleFactor):

    # Determine what models are available
    modelFactors = listSuperResModels(debug=False)
    possibleModels = list(modelFactors.keys())

    if not modelName.lower() in [p.lower() for p in possibleModels]:
        print(f'Error: requested model \'{modelName}\' not found!')
        return None

    # We could just put the keys into the dict in lowercase, but that would mean that they
    # would be displayed like that in the help screen, so I'd rather just it like this.
    for key in possibleModels:
        if modelName.lower() == key.lower():
            # Make sure the requested scaling factor is available
            if not upscaleFactor in modelFactors[key]:
                print(f'Error: requested upscale factor ({upscaleFactor}) is not available for model \'{modelName}\'! \nAvailable factors are: {modelFactors[key]}')
                return None

            # Otherwise, we are good to go, and can take this file as our model
            properName = modelName.lower()
            modelPath = MODELS_DIR + f'{key}/{key}_x{upscaleFactor}.pb'
            break

    # Since fsrcnn has a small option, which is just a smaller network, but the same structure, upscale
    # we need to adjust the proper name to just fsrcnn
    if properName == 'fsrcnn_small':
        properName = 'fsrcnn'
    
    # Create the network
    model = dnn_superres.DnnSuperResImpl_create()
    # Read in the .pb file containing the weights
    model.readModel(modelPath)
    # Set the active model type to the one we just read in
    model.setModel(properName, upscaleFactor)

    return model


def upsample(image, upscaleFactor=defaultModelFactor, modelName=defaultModelType):
    global defaultModelInstance, defaultModelType, defaultModelFactor
    # Make sure we have a (singleton) model on hand
    if defaultModelInstance is None or upscaleFactor != defaultModelFactor or modelName != defaultModelType:
        defaultModelInstance = prepareSuperResModel(modelName, upscaleFactor)
        defaultModelType = modelName
        defaultModelFactor = upscaleFactor
   
    if defaultModelInstance is None:
        return None

    # Now do the upscaling
    return defaultModelInstance.upsample(image)

def downsample(image, downscaleFactor=2):
    """
    A very simple wrapper around either `cv2.pyrDown()` or
    `skimage.measure.block_reduce()`, depending on the desired
    downscale factor.

    Parameters
    ----------

    image : np.ndarray[H,W]
        The image to be downsampled

    downscaleFactor : int
        The factor by which to downsample the image.

    Returns
    -------

    dsImage : np.ndarray[H/downscaleFactor,W/downscaleFactor]
        Downsampled image.
    """
    if downscaleFactor == 2:
        return cv2.pyrDown(image)
    else:
        return block_reduce(image, downscaleFactor)
    
